import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import javax.swing.JFrame; 
import javax.swing.ImageIcon; 

import com.shigeodayo.pframe.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class NLEDMatrixPatcher extends PApplet {

/*
Copyright (c) 2016 Northern Lights Electronic Design, LLC
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 
 Original Author: Jeffrey Nygaard
 Company: Northern Lights Electronic Design, LLC
 Contact: JNygaard@NLEDShop.com
 Date Updated: August 26, 2016 
 Software Version:  1f
 Webpage: www.NLEDShop.com/nledpatcher
 Written in Processing v2.2.1(won't work with 3.2.1)  - www.Processing.org
 
 //============================================================================================================
 
 This software is written to work with NLED-Matrix, the free Open-Source, Processing/Java based, LED Matrix Control software
 And works with NLED Aurora - Color Sequence Creation Software
 
 Used to create a coordinate Patch file usable with NLED-Matrix,NLED Aurora, NLED Live Patcher. Required for custom layouts and control schemes for LED Panels/Matrices.
 
 Feel free to use this software or parts of it for whatever you like. Personal or commercial, just give attribution if the customized project is released.
 And Please Contact Me at the email address above with any questions, comments, or suggestions.
 
 See included README.TXT for more details
 
 //============================================================================================================
 
 This software opens 2 windows, one window is for the controls the other is the grid window that displays the layout.
 
 Control Window Usage:
 
 	 - Figure out how many pixels wide and how many pixels tall the matrix will be. Figure out which order the LEDs are arranged in.
 	 
 	 - Select the amount of Cells for how many grids wide the Grid Window will display as a square.
 	 
 	 - Adjust the size of the window by dragging an outside edge to fit your screen
 	 
 	 - Adjust the Cell Size to suit or click the Fit Grid To Window button.  Click Update Grid to apply changes.
 	 
 	 - Enter the total amount of pixels in the LED matrix using the slider or textfield
 	 
 === Switch to the Grid window ===
 	 
 	 - Start placing the points one by one, points can be moved by clicking, holding and then dragging the icon to new location. 
 	 
 	 - Once placed they can be moved by clicking and dragging the icon.
 	 
 	 - For faster placement, in the Control Window press the Drag To Place button(or press Shift to toggle). Then back in the Grid Window, click, hold and move the mouse to place
 	    the pixel icons. Note you must disable Drag To Place to restore normal functions.
 
 	- It is recommended to save inclemently often, some functions can not be undone. The software automatically saves before functions run, press "CTRL z" to attempt undo
 
 Controls:
 
 	- At any time Right-Click a LED Icon to open a options menu.
 		  - Delete Before Pixel Number, deletes all pixels before selected number and placing points will begin at the lowest pixel that was deleted, then
 			  once it reaches the already placed pixels, it will begin at the highest pixel that hasn't been placed yet
 		  - Delete After Pixel Number
 		  - Close Menu
 		
         - At anytime Right-Click an empty grid square to open the Tools Menu
           - Group Select, after selection press again to start bounding box drag, select the icons/points to be grouped
           - Move Group Selection, once clicked, clicking again will begin a rectangle drag that will select all pixels within it, when released the pixels will move with the mouse, until mouse press which places them.
           - Copy Group, copies the group into the next lowest ID#s, starts Move Group once the points are created
           - Delete Group, deletes the currently selected points. After deletion when points are placed they will be placed into the lowest ID#s
           - Rotate Group CW - Rotate Group Clockwise 90 degrees
           - Rotate Group CCW - Rotate Group Counter Clockwise 90 degrees
           
 	- Use the Up/Down/Left/Right Arrow keys to move the display around on the grid window. Or use the X/Y sliders on the Control Window or Grid Window
 
 	- CTRL+Z loads previous autosave, limited functionality
 
 	- Shift key Toggles Drag to Place usage
 
	- Spacebar simulates a Right-Mouse click to open the menu, for Mac users
 
 =============================================================================================================================================
 
 Revision 1a to 1b:
   	- Added rolling autosaves
   	- Added save/restore window saves
   	- Added tallys for rows and columns to count how many points are in them	
   	- Added some hotkeys
   	- Added scroll bars on Grid Window, dragging them causes some graphical glitches though otherwise use control window sliders or arrow keys	
   	- Fixed dragging points	
   	- Added option to load an image to overlay the points onto. Can't resize or move it, so do that all in a image editor before loading. Doesn't save image path.
   
 Revision 1b to 1c:
   -Updated Group Moving
   -Updated draw method, by removing loop() and replacing with redraw()
   	
 Revision 1c to 1d:
   - Fixed some minor bugs
   - Fixed grid layout and grid numbering layout issues
   
 Revision 1d to 1e:
   - Revised Menus
   - Added Copy Group
   - Added Rotate Group, Clockwise(CW) or Counter Clockwise(CCW)
   - Revised how points/icons can be placed, more dynamic.
   - Linked CTRL+Z usage with autosaves, easier and more capable than before. Note it saves all points even if they are not Enabled(not seen) but have yet to be replaced.
       As a work around, finish the patch file in one session, or move points to empty space when saving if planning on finishing later.
   
  Revision 1e to 1f:
   - Revised Menus
   - Added Copy Group
   - Added Rotate Group, Clockwise(CW) or Counter Clockwise(CCW)
   - Revised how points/icons can be placed, more dynamic.
   - Linked CTRL+Z usage with autosaves, easier and more capable than before. Note it saves all points even if they are not Enabled(not seen) but have yet to be replaced.
       As a work around, finish the patch file in one session, or move points to empty space when saving if planning on finishing later.
     
 Notes:

   - Rotating while over other lapping other points causes the points to not get rotated correctly, 
   for now rotate so points would not be rotated onto others.
   
   - LEDGraphics Object only has up to 4096 objects, so that is maximum amount of pixels, but the number could be increased in the declaration.
   
 Updates:
 
   - Make drag to place change mouse cursor
   
   - Add button to step through LED Icons changing their color to check for mistakes(like I made)
  
   - Loading a file with lower amount of placed icons, won't wipe them but allow them to be moved. need to wipe clean
   
   - Current export doesn't load into matrix without deleteing new header info and leaving just maxpixels
 
 */

//import libraries


java.awt.Insets insets;

//Init Constants
final int ColButtonsBG = color(100);
final int ColButtonsHL = color(200, 40, 40);
final int ColButtonsOut = color(0, 100, 255); //selected/highlighted color for buttons
final int LEDIconColor = color(0, 0, 255);
final int cBackgroundColor = color(65);
final int cCellSpacing = 22;

final int cMaxAutoSaves = 10;
final boolean cAutoSaveEnabled = true;

//============= VARIABLES ==================================
int OffsetVarX = 0;
int OffsetVarY = 0;

int CellSpacing = 0;
int IconCounter = 0;
int ChannelNum = 0;
int cellWidth = 32;

int PrevHeight = height;  //used to watch if the grid window's size is changed
int PrevWidth = width;
int UserCellSpacing = cCellSpacing;

int CoordX = 0;
int CoordY = 0;

int SliderSet=0;
int SelectedSlider=0;
int SelectedTextField = 0;
int SelectedPoint = 0;
int FileBrowserAction = 0;

int MenuOpenedIcon = 0;

int MenuXpos = 0;
int MenuYpos = 0;

int SelectBoxX = 0;
int SelectBoxY = 0;

int SelectBoxXpos = 0;
int SelectBoxYpos = 0;

int SelectBoxW = 0;
int SelectBoxH = 0;

int GroupedPoints[] = new int[1];
int SelectedPointsCounter = 0;

int LowestX = 10000;
int LowestY = 10000;

int rollingSaveNum = 1;

int tempTally = 0;

boolean GlobalDragging = false;
boolean DraggingPoint = false;    
boolean DragToDraw=false;
boolean TextFieldActive = false;
boolean DragToPlaceEnabled = false;
boolean MenuOpen = false;
boolean ToolsMenuOpen = false;
boolean EnableGroupSelect = false;
boolean SelectPointInit = false;
boolean DraggingGroup = false;
//boolean CopyGroup = false;
boolean DeleteGroup = false;
//boolean AllowAutoSave = true;

PImage UserImage;

String SelectedFile ="";     

int ColorArray[] = new int[100];
int CoordColor = color(0, 0, 255);

ArrayList strs = new ArrayList();
ArrayList groupSelAL = new ArrayList();

//================ Object Declaration ========================
LEDgraphic[] LEDIcons;
Button[] Buttons;
SliderBar[] SelectSliders;
TextField[] TextFields;

SliderBar GridXScroll;
VertSliderBar GridYScroll;

PFrame f; //second window objects
secondApplet s;  //code taken from processing forum

public void setup() 
{
  size(600, 600);
  background(255);
  noLoop();

  PFrame f = new PFrame();
  LEDIcons = new  LEDgraphic[4096];

  LEDIcons[0] = new LEDgraphic(0, 0, 0, 0, 0,0);

  frame.setResizable(true);
  ImageIcon titlebaricon = new ImageIcon(loadBytes("favicon.gif"));
  frame.setIconImage(titlebaricon.getImage());
  frame.setTitle("Grid Window - NLED Matrix Patcher v.1f");
  f.setIconImage(titlebaricon.getImage());
  f.setTitle("Control Window - NLED Matrix Patcher v.1f");

  CellSpacing = cCellSpacing;//width / cellWidth;
  PrevWidth = width;
  PrevHeight = height;

  insets = frame.getInsets();	//get title bar and border sizes to adjust size correctly

  GridXScroll = new SliderBar(0, height-20, width-20, 20, 0, 0, cellWidth, color(65), color(0, 175, 235), color(65), color(0), true, false); //X
  GridYScroll = new VertSliderBar(width-20, 0, 20, height-20, 0, 0, cellWidth, color(65), color(0, 175, 235), color(65), color(0), true, false); //Y
}//end setup


public void draw() 
{
  background(255);
  //adjust window position based on X/Y sliders
  translate(OffsetVarX*CellSpacing, OffsetVarY * CellSpacing);

  if (UserImage != null) image(UserImage, CellSpacing*2, CellSpacing*2);

  if (PrevWidth != width || PrevHeight !=height)
  {
    //check if grid window has been resized, replace coords if it has
    PrevWidth = width;
    PrevHeight = height;   
    println("Width Changed"); 
    CellSpacing= UserCellSpacing;//width / (cellWidth+1);

    //adjust grid window sliders
    GridXScroll.ypos = height-20;
    GridXScroll.Slwidth = width-20;

    GridYScroll.xpos = width-20;
    GridYScroll.Slheight = height-20;

    //move all the LEDIcons into new grid size
    for (int i=0; i != IconCounter; i++)
    {   
      LEDIcons[i].xpos = (LEDIcons[i].xCoord * CellSpacing)+(CellSpacing/2);
      LEDIcons[i].ypos = (LEDIcons[i].yCoord * CellSpacing)+(CellSpacing/2);    
      LEDIcons[i].bSize = (int)(CellSpacing*0.85f);
    }
  }//end resize If

  //====================== REDRAW GRID AND LABELS ======================
  stroke(0); //line stroke color black
  strokeWeight(1);

  int holdi=0;

  for (int i=0; i != cellWidth+1; i++)
  {
    line(CellSpacing, (CellSpacing*2)+(i*CellSpacing), (CellSpacing)*(cellWidth+3), (CellSpacing*2)+(i*CellSpacing)); //horiz grid lines
    line((CellSpacing*2)+(i*CellSpacing), CellSpacing, (CellSpacing*2)+(i*CellSpacing), CellSpacing*(cellWidth+2)); //vert grid lines

    holdi=i;

    if (i != cellWidth)
    {
      textAlign(CENTER);  
      textSize(CellSpacing/2);
      fill(0);
      text(i+1, (CellSpacing*2)-(CellSpacing/2), (CellSpacing*2)+(i*CellSpacing)+(CellSpacing/2)); //left side numbers
      text(i+1, (CellSpacing*2)+(i*CellSpacing)+(CellSpacing/2), (CellSpacing*2)-(CellSpacing/2)); //top numbers 
      text(i+1, (CellSpacing*2.5f)+(cellWidth*CellSpacing), (CellSpacing*2)+(i*CellSpacing)+(CellSpacing/2)); //right side numbers
      text(i+1, (CellSpacing*2)+(i*CellSpacing)+(CellSpacing/2), (CellSpacing*(cellWidth+3))-(CellSpacing/2)); //bottom numbers

      //START TALLY	
      for (int q=0; q != IconCounter; q++)
      {   
        if (LEDIcons[q].yCoord == i+2) tempTally++;
      }  
      fill(0, 0, 255);  //fill blue
      text(tempTally, CellSpacing/2, (CellSpacing*3)+((i-1)*CellSpacing)+(CellSpacing/2)); //left side tally numbers 
      tempTally = 0;

      //tally column

      for (int q=0; q != IconCounter; q++)
      {   
        if (LEDIcons[q].xCoord == i+2) tempTally++;
      }  

      fill(0, 0, 255);  //fill blue
      text(tempTally, (CellSpacing*3)+((i-1)*CellSpacing)+(CellSpacing/2), CellSpacing/2); //top numbers 		
      tempTally = 0;
    }	//end if	
    //end tally
  }
  noStroke();
  //=====================  END DRAW GRID ======================

  for (int q=0; q != SelectedPointsCounter; q++)  LEDIcons[GroupedPoints[q]].displaySelected(); //displays red border, indicating selected
  for (int i=0; i != IconCounter; i++) if (LEDIcons[i].Enabled == true) LEDIcons[i].display();  //displays icons

  //==================================================================

  if (DragToDraw == true)
  {
    PlacePoint();
    return;
  }

  //check if dragging and display
  if (DraggingGroup == true && MenuOpen == false && ToolsMenuOpen == false)
  {
    for (int q=0; q != SelectedPointsCounter; q++)
    {  
      LEDIcons[GroupedPoints[q]].xpos =  mouseX + ((LEDIcons[GroupedPoints[q]].xCoord*CellSpacing) - LowestX) + (LEDIcons[GroupedPoints[q]].bSize/2);
      LEDIcons[GroupedPoints[q]].ypos =  mouseY + ((LEDIcons[GroupedPoints[q]].yCoord*CellSpacing) - LowestY) + (LEDIcons[GroupedPoints[q]].bSize/2);
    }
  } else if (DraggingPoint == true) //if dragging, display drag and set draw() to loop til done dragging
  {  
    //factors in translate()
    LEDIcons[SelectedPoint].xpos = mouseX-(OffsetVarX*CellSpacing);
    LEDIcons[SelectedPoint].ypos = mouseY -(OffsetVarY * CellSpacing);
  }
  //end dragging

    if (MenuOpen == true)
  {
    textAlign(LEFT); 
    fill(100);
    rect(MenuXpos, MenuYpos, 160, 75);
    fill(70);    
    rect(MenuXpos, MenuYpos+25, 160, 25); 
    // rect(MenuXpos, MenuYpos+75, 160, 25);  

    fill(255);
    textSize(12);
    text("Delete Before Point "+(MenuOpenedIcon+1), MenuXpos+10, MenuYpos+15);     
    text("Delete After Point "+(MenuOpenedIcon+1), MenuXpos+10, MenuYpos+40);
    text("Close Menu", MenuXpos+10, MenuYpos+65);
  } else if (ToolsMenuOpen == true)
  {
    textAlign(LEFT); 
    fill(100);
    rect(MenuXpos, MenuYpos, 160, 175); //was 125
    fill(70);    
    rect(MenuXpos, MenuYpos+25, 160, 25); 
    rect(MenuXpos, MenuYpos+75, 160, 25);  
    rect(MenuXpos, MenuYpos+125, 160, 25);      
    fill(255);
    textSize(12);
    text("Group Select", MenuXpos+10, MenuYpos+15);     
    text("Move Group", MenuXpos+10, MenuYpos+40);
    text("Copy Group", MenuXpos+10, MenuYpos+65);
    text("Delete Group", MenuXpos+10, MenuYpos+90);   
    text("Rotate Group CW", MenuXpos+10, MenuYpos+115); 
    text("Rotate Group CCW", MenuXpos+10, MenuYpos+140);    
    text("Close Menu", MenuXpos+10, MenuYpos+165);
  }

  //Draws Selection Box
  if (EnableGroupSelect == true && SelectPointInit == true) 
  {
    fill(0, 0, 0, 0);
    strokeWeight(2);
    stroke(255, 0, 0);

    int TempX = 0;
    int TempY = 0;     
    int TempXb = 0;
    int TempYb = 0;   

    if (SelectBoxX > mouseX) {
      TempX = mouseX;//SelectBoxX;
      TempXb = SelectBoxX - mouseX;
      SelectBoxW = TempXb;
      SelectBoxXpos = TempX;
    } else {
      TempX = SelectBoxX;
      TempXb =  mouseX - SelectBoxX;
      SelectBoxW = TempXb;
      SelectBoxXpos = SelectBoxX;
    }  
    if (SelectBoxY > mouseY) {
      TempY = mouseY;
      TempYb =   SelectBoxY - mouseY;
      SelectBoxH = TempYb;   
      SelectBoxYpos = TempY;
    } else {
      TempY =  SelectBoxY;
      TempYb = mouseY - SelectBoxY;
      SelectBoxH = TempYb;
      SelectBoxYpos = SelectBoxY;
    }  
    stroke(0, 255, 0); //draw selection rectangle
    rect(TempX, TempY, TempXb, TempYb);
  }

  //translations and scrollbar displays last
  translate(-(OffsetVarX*CellSpacing), -(OffsetVarY * CellSpacing));

  GridXScroll.displayOnGridWindow();
  GridYScroll.display();
}//end draw() func

//==================================================================================

public void PlacePoint()
{
  int testX = 0;
  int testY = 0;

  CoordX = round((mouseX / CellSpacing))-OffsetVarX;
  CoordY = round((mouseY / CellSpacing))-OffsetVarY;
  if (CoordX < 2 || CoordY < 2) return; //don't place on outer edge with labels

  println("PlacePoint() with "+CoordX+", "+CoordY);

  //run autosave function now before changes
  if (cAutoSaveEnabled == true && DragToPlaceEnabled == false)    RunRollingAutoSave(); //if not the other methods do autosave.

  if (DraggingGroup == true)
  {
    DraggingGroup = false;
    //check if any overlap
    for (int x=0; x != SelectedPointsCounter; x++)     
    {
      SelectedPoint = GroupedPoints[x];  
      CoordX = round((LEDIcons[SelectedPoint].xpos / CellSpacing))+OffsetVarX;
      CoordY = round((LEDIcons[SelectedPoint].ypos / CellSpacing))+OffsetVarY;   

      for (int q=0; q != IconCounter; q++) 
      {    
        boolean TempFlag = false;
        //is q == to any of the IDs on the list?
        //set the flag if it is
        for (int n=0; n != SelectedPointsCounter; n++) 
        {
          if (GroupedPoints[n] == q) TempFlag = true;
        }//end n for()

        if (LEDIcons[q].over(CoordX, CoordY) && TempFlag == false) //check if over() and exclude the points being dragged from being checked
        {
          //if a dragging point is over another point, even itself
          //there was a point being dragged that attempted to be dropped onto of a non-dragging point
          println("Can Not Drop, Point Already There "+q+" :"+SelectedPoint);

          DraggingGroup = true;
          redraw();
          return;
        }//end over() if()
      } //end q for()
    }//end x for()

    //no points overlap 
    for (int q=0; q != SelectedPointsCounter; q++)
    {  
      SelectedPoint = GroupedPoints[q];    
      CoordX = round((LEDIcons[SelectedPoint].xpos / CellSpacing))+OffsetVarX;
      CoordY = round((LEDIcons[SelectedPoint].ypos / CellSpacing))+OffsetVarY; 

      testX = (CoordX * CellSpacing)+(CellSpacing/2);
      testY = (CoordY * CellSpacing)+(CellSpacing/2);
      
      LEDIcons[SelectedPoint].xpos = testX;
      LEDIcons[SelectedPoint].ypos = testY;
      LEDIcons[SelectedPoint].xCoord = CoordX;
      LEDIcons[SelectedPoint].yCoord = CoordY;
  } 
    println("Group Placed");
    noLoop();
    return;
  } //end if DraggingGroup

  //else place a point

  fill(CoordColor); //fill the coordinate with color, defined at top

  //println("X: " +mouseX+" - Y: "+mouseY); //make shows prints X - Y coordinates  

  CoordX = round((mouseX / CellSpacing))-OffsetVarX;
  CoordY = round((mouseY / CellSpacing))-OffsetVarY;

  testX = (CoordX * CellSpacing)+(CellSpacing/2);
  testY = (CoordY * CellSpacing)+(CellSpacing/2); 

  //duplicated for use with delete before functions  
  for (int q = 0; q != IconCounter; q++)  
  {
    if (LEDIcons[q].over(CoordX, CoordY) && LEDIcons[q].Enabled == true)
    {
      if (DragToDraw == true) return; // don't allow a move point if Drag Drawing
      println("Point Already There");
      DraggingPoint = true;
      SelectedPoint = q;
      loop();
      return;
    }
  } //end for()
  //end duplication

    //println("Coord X: "+round((mouseX / CellSpacing))+" Y: "+round((mouseY / CellSpacing)));

  redraw(); //displays new point, do it here in case of return

  for (int q = 0; q != IconCounter; q++)  
  {
    if (LEDIcons[q].Enabled == false)
    {
      LEDIcons[q] = new LEDgraphic(q, testX, testY, CoordX, CoordY, (int)(CellSpacing*0.85f));
      LEDIcons[q].display();
      return;
    } //end if()
  } //end for() 

  if (IconCounter >= SelectSliders[0].Value)
  {
    println("Maximum LED Amount Reached"); 
    redraw();
    return;
  }

  //else never found an open ID#, create new one
  LEDIcons[IconCounter] = new LEDgraphic(IconCounter, testX, testY, CoordX, CoordY, (int)(CellSpacing*0.85f));
  LEDIcons[IconCounter].display();
  IconCounter++;
} //end func

//==================================================================================

public void UpdateGridCells()
{
  cellWidth = SelectSliders[2].Value;
  CellSpacing = SelectSliders[1].Value;
  UserCellSpacing= SelectSliders[1].Value;

  SelectSliders[3].Maximum = cellWidth;
  SelectSliders[4].Maximum = cellWidth;         

  GridXScroll.Maximum = cellWidth/2;
  GridYScroll.Maximum = cellWidth/2;

  println("SCROLL SET TO MAX: "+cellWidth);

  for (int q=0; q != IconCounter; q++)
  {   
    LEDIcons[q].xpos = (LEDIcons[q].xCoord * CellSpacing)+(CellSpacing/2);
    LEDIcons[q].ypos = (LEDIcons[q].yCoord * CellSpacing)+(CellSpacing/2);    
    LEDIcons[q].bSize = (int)(CellSpacing*0.85f);
  }

  OffsetVarX = 0;
  OffsetVarY = 0;         
  ForceRefreshGrid(); //would just want to do a redraw() on the grid window but it doesn't work
}

//==================================================================================

public void ForceRefreshGrid()
{
  println("Grid Window Refreshed");
  //seems have to do it this way in order to call redraw from the Control Window
  redraw();
}

//functions and setup for Control Window

public class PFrame extends JFrame {
	public PFrame() {
		setBounds(100, 100, 500,400);
		s = new secondApplet();
		add(s);
		s.init();
		show();
		this.setResizable(false);
	}
}

public class secondApplet extends PApplet {
	public void setup() {
		// background(127);
		Buttons = new Button[11];
		Buttons[0] = new Button("Save File", 10, 40, 120, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);
		Buttons[1] = new Button("Load File", 140, 40, 120, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);
		Buttons[2] = new Button("Update Grid", 170, 230, 130, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);       
		Buttons[3] = new Button("Reset Icons", 10, 310, 130, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);    
		Buttons[4] = new Button("Reset Position", 10, 270, 130, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);    
		Buttons[5] = new Button("Fit Grid To Window",310, 230, 170, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);    
		Buttons[6] = new Button("Drag to Place", 340, 40, 140, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, true, false);    
		Buttons[7] = new Button("EULA", 425, 10, 60, 20, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);    
		Buttons[8] = new Button("Website",10, 10, 60, 20, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);  
		//new
		Buttons[9] = new Button("Save Patch",10, 80, 120, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);  
		Buttons[10] = new Button("Load Image",140, 80, 120, 30, ColButtonsBG, ColButtonsHL, ColButtonsOut, false, false);  
		
		SelectSliders = new SliderBar[5];
		SelectSliders[0] = new SliderBar(140, 120, 230, 20, 8, 0, 340, color(200), color(0), ColButtonsOut, color(0), true, false); //channels
		SelectSliders[1] = new SliderBar(140, 160, 230, 20, cCellSpacing, 0, 80, color(200), color(0), ColButtonsOut, color(0), true, false); //cell size
		SelectSliders[2] = new SliderBar(140, 200, 230, 20, 32, 0, 256, color(200), color(0), ColButtonsOut, color(0), true, false); //cells

		SelectSliders[3] = new SliderBar(260, 280, 200, 20, 0, 0, 32, color(255), color(0), color(200, 200, 200), color(0), true, false); //X
		SelectSliders[4] = new SliderBar(260, 310, 200, 20, 0, 0, 32, color(255), color(0), color(200, 200, 200), color(0), true, false); //Y
		
		TextFields = new TextField[3];
		TextFields[0] = new TextField("Enter #", 380, 120, 80, 25, color(255), color(200, 40, 40), true, false);
		TextFields[1] = new TextField("Enter #", 380, 160, 80, 25, color(255), color(200, 40, 40), true, false);
		TextFields[2] = new TextField("Enter #", 380, 200, 80, 25, color(255), color(200, 40, 40), true, false);

		frameRate(10);
	}

	public void draw() 
	{
		background(cBackgroundColor);
		fill(255);
		textSize(18);
		textAlign(LEFT); 
		text("NLED Matrix Patcher Software v.1f", 100, 20);   
		textSize(14);  

		text("X Position: "+ SelectSliders[3] .Value, SelectSliders[3].xpos -100, SelectSliders[3].ypos+(SelectSliders[3].Slheight)); 
		text("Y Position: "+ SelectSliders[4] .Value, SelectSliders[4].xpos -100, SelectSliders[4].ypos+(SelectSliders[4].Slheight)); 
		
		textAlign(RIGHT);  
		text("Total Pixels: "+SelectSliders[0].Value, SelectSliders[0].xpos -20, SelectSliders[0].ypos+(SelectSliders[0].Slheight));
		text("Cell SIze: "+SelectSliders[1].Value, SelectSliders[1].xpos -20, SelectSliders[1].ypos+(SelectSliders[1].Slheight)); 
		text("Cells: "+SelectSliders[2].Value, SelectSliders[2].xpos -20, SelectSliders[2].ypos+(SelectSliders[2].Slheight)); 

		textAlign(CENTER);
		text("Type in Value\n& Press Enter:", 420, 90);
		text("www.NLEDShop.com/nledpatcher      www.NLEDShop.com/nledmatrix", 250, 360);
		textAlign(LEFT);   
		text("Icons Placed: "+IconCounter,10,250);

		for (int i=0;i != Buttons.length;i++) Buttons[i].display();
		for (int i=0;i !=  TextFields.length;i++)  TextFields[i].display();
		for (int i=0;i !=  SelectSliders.length;i++)  SelectSliders[i].display();

		if (GlobalDragging==true) //only check sliders if one is being dragged
		{
			switch(SliderSet)
			{
			case 0:
				if (SelectSliders[SelectedSlider].Dragging==true) //only display while Dragging is set
				{ 
					SelectSliders[SelectedSlider].move(mouseX, mouseY); 
					SelectSliders[SelectedSlider].display();

					if (SelectedSlider == 3) 
					{
						OffsetVarX = -SelectSliders[SelectedSlider].Value;
						ForceRefreshGrid();
					}
					if (SelectedSlider == 4) 
					{
						OffsetVarY = -SelectSliders[SelectedSlider].Value;
						ForceRefreshGrid();
					}          
				}   
				break;
			}//end switch
		}//end if
	}//end draw

	public void fileSelected(File selection) 
	{
		if (selection == null) {
			println("Window was closed or the user hit cancel.");
		} 
		else {
			println("User selected " + selection.getAbsolutePath());
			SelectedFile = selection.getAbsolutePath();

			switch(FileBrowserAction)  
			{
			case 0: //generate
				WriteCustomPatchFile();
				break;
			case 1: //save Layout
				SaveLayoutFile();
				break;
			case 2: //Load Layout
				LoadLayoutFile();
				break;
			case 3: //Load Image
				UserImage = loadImage(SelectedFile);
				ForceRefreshGrid(); //would just want to do a redraw() on the grid window but it doesn't work   
				break;		
			}//end switch
		}//end else
	}//end func

	public void mouseReleased()
	{
		if (GlobalDragging==true)
		{
			GlobalDragging=false;

			println("Slider Released");
			switch(SliderSet)
			{
			case 0:
				SelectSliders[SelectedSlider].Dragging=false;
				break;
			}
		}
	}

	public void mousePressed() 
	{
		println("Second Window - X: " +mouseX+" - Y: "+mouseY); //make shows pritns X - Y coordinates  

		for (int q=0; q != SelectSliders.length;q++)  
		{
			if (SelectSliders[q].over(mouseX, mouseY))
			{
				println("Selected Slider: "+q);
				SliderSet=0;
				SelectedSlider=q;
				SelectSliders[q].Dragging=true;
				GlobalDragging=true;
			}
		}

		for (int i=0;i != TextFields.length;i++)
		{
			if (TextFields[i].over(mouseX, mouseY))
			{
				TextFields[SelectedTextField].Selected = false;
				SelectedTextField = i;
				TextFieldActive = true;
				TextFields[SelectedTextField].Label = "";
				TextFields[SelectedTextField].Selected = true;
				println("Text Field "+i);
			}
		}

		for (int i=0;i != Buttons.length;i++)
		{
			if (Buttons[i].over(mouseX, mouseY))
			{
				//certain buttons if already selected, stay selected(increment, all)
				Buttons[i].Selected=! Buttons[i].Selected;

				println("Button"+i);
				switch(i)
				{
				case 0: //save
					FileBrowserAction = 1;
					selectOutput("Select a file to Save To:", "fileSelected");
					break;
				case 1: //load
					FileBrowserAction = 2;
					selectInput("Select a file to Load:", "fileSelected");
					break;      
				case 2: //Update Cells
					UpdateGridCells();
					break;   
				case 3://Reset Icons
					IconCounter = 0;
					ForceRefreshGrid(); //would just want to do a redraw() on the grid window but it doesn't work     
					break;
				case 4://Reset Position
					OffsetVarX = 0;
					OffsetVarY = 0;         
					ForceRefreshGrid(); //would just want to do a redraw() on the grid window but it doesn't work     
					break;          
				case 5://Fit to Window
					//so get the size of the grid window, divide it by cellwidth or something
					SelectSliders[1].Value = (PrevWidth-40)/(cellWidth+2);
					//println(PrevWidth/cellWidth);
					UpdateGridCells();
					break;            
				case 6://Drag to Place Toggle
					if(Buttons[i].Selected == true)  DragToPlaceEnabled = true;
					else   DragToPlaceEnabled  = false;
					break;
				case 7: //open EULA
					link("http://creativecommons.org/licenses/by/4.0/legalcode");
					break;
				case 8: //open Website
					link("http://www.NLEDshop.com/nledmatrix");
					break; 
				case 9: //save Patch
					FileBrowserAction = 1;
					selectOutput("Select a file to Save To:", "fileSelected");
					break; 
				case 10: //load Image
					FileBrowserAction = 3;
					selectInput("Select a file to Load:", "fileSelected");
					break;   
				}//end switch
			}//end if
		}//end for
	}//end mousepressed

	//================================================================================== 

	public void keyPressed()
	{
		println(PApplet.parseInt(key)); 

		if(TextFieldActive == true)
		{
			if(key == ENTER || key == RETURN)
			{
				TextFieldActive = false;
				TextFields[SelectedTextField].Selected = false;   
				if(PApplet.parseInt(TextFields[SelectedTextField].Label) > SelectSliders[SelectedTextField].Maximum)SelectSliders[SelectedTextField].Maximum = PApplet.parseInt(TextFields[SelectedTextField].Label);
				SelectSliders[SelectedTextField].Value =  PApplet.parseInt(TextFields[SelectedTextField].Label);
				return;
			}
			else if (key=='1' || key== '2' || key=='3' || key=='4' || key=='5' || key=='6' || key=='7' || key=='8' || key=='9' || key=='0')
			TextFields[SelectedTextField].Label =  TextFields[SelectedTextField].Label + PApplet.parseChar(key);
			TextFields[SelectedTextField].display();
			return;
		}
	} 

} //end

public void RotateGroup(int passedAngle)
{
	println("RotateGroup() with "+passedAngle);
/*
//sample code from internet
float x_old = p.x; float y_old = p.y;
p.x = x_old * cos(a) - y_old * sin(a);
p.y = x_old * sin(a) + y_old * cos(a);

float tempFloatX = LEDIcons[q].xCoord; 
float tempFloatY = LEDIcons[q].yCoord;
float a = 1.5708;

LEDIcons[q].xCoord = tempFloatX * cos(a) - tempFloatY * sin(a);
LEDIcons[q].yCoord = tempFloatX * sin(a) + tempFloatY * cos(a);
*/

	LowestX = 10000;
	LowestY = 10000; 

	float tempFloatX = 0;
	float tempFloatY = 0;

	float a = radians(passedAngle);//1.5708;

	float cosVal = cos(a); //save values since they stay the same
	float sinVal = sin(a);

	for (int q=0; q != SelectedPointsCounter; q++) 
	{   
		tempFloatX = LEDIcons[GroupedPoints[q]].xCoord;
		tempFloatY = LEDIcons[GroupedPoints[q]].yCoord;

		//println("START : "+GroupedPoints[q]+" : "+LEDIcons[GroupedPoints[q]].xCoord+", "+LEDIcons[GroupedPoints[q]].yCoord);

		//LEDIcons[GroupedPoints[q]].xCoord = int((float)(tempFloatX * cos(a) - tempFloatY * sin(a)));
		//LEDIcons[GroupedPoints[q]].yCoord = int((float)(tempFloatX * sin(a) + tempFloatY * cos(a)));
		if (passedAngle == 90)
		{
			LEDIcons[GroupedPoints[q]].xCoord = PApplet.parseInt(ceil(tempFloatX * cosVal - tempFloatY * sinVal));
			LEDIcons[GroupedPoints[q]].yCoord = PApplet.parseInt(ceil(tempFloatX * sinVal + tempFloatY * cosVal));
		} 
		else if (passedAngle == 270)
		{
			LEDIcons[GroupedPoints[q]].xCoord = PApplet.parseInt(floor(tempFloatX * cosVal - tempFloatY * sinVal));
			LEDIcons[GroupedPoints[q]].yCoord = PApplet.parseInt(floor(tempFloatX * sinVal + tempFloatY * cosVal));
		} 
		else println("ERROR WRONG ANGLE ENTERED");

		//println("END: "+LEDIcons[GroupedPoints[q]].xCoord+", "+LEDIcons[GroupedPoints[q]].yCoord);

		if (LEDIcons[GroupedPoints[q]].xCoord < LowestX)  LowestX = LEDIcons[GroupedPoints[q]].xCoord;
		if (LEDIcons[GroupedPoints[q]].yCoord < LowestY)  LowestY = LEDIcons[GroupedPoints[q]].yCoord;
		//its dragging which will update the xpos and ypos
	}

	//println(LowestX+" : "+LowestY);

	LowestX = LowestX*CellSpacing; //set new lowest values
	LowestY = LowestY*CellSpacing;

	DraggingGroup = true;
	loop();
} //end func 

//==================================================================================

public void InitMoveGroup()
{
	LowestX = 10000;
	LowestY = 10000; 
	for (int q=0; q != SelectedPointsCounter; q++) 
	{  
		if (LEDIcons[GroupedPoints[q]].xpos < LowestX)  LowestX = LEDIcons[GroupedPoints[q]].xpos;
		if (LEDIcons[GroupedPoints[q]].ypos < LowestY)  LowestY = LEDIcons[GroupedPoints[q]].ypos;
	}    
	DraggingGroup = true;
	loop(); //start looping draw() til icons are placed
}

//==================================================================================

public void RunCopyGroup()
{
	println("RunCopyGroup()");
	//lowest ID is the first to get a duplicated
	int testX = 0;
	int testY = 0;
	int i =0;

	int CountPlaced = 0;
	
	// for (int q=0; q != IconCounter; q++) println(LEDIcons[q].Enabled);

	for (int q=0; q != SelectedPointsCounter; q++)
	{    
		testX = (LEDIcons[GroupedPoints[q]].xCoord * CellSpacing)+(CellSpacing/2);
		testY = (LEDIcons[GroupedPoints[q]].yCoord * CellSpacing)+(CellSpacing/2);   

		//close needs an else, how to make a new one if none is found using icon counter?
		for (i = 0; i != IconCounter; i++)  
		{

			if (LEDIcons[i].Enabled == false)
			{
				println("R: "+i+" and "+q);
				LEDIcons[i] = new LEDgraphic(i, testX, testY, LEDIcons[GroupedPoints[q]].xCoord, LEDIcons[GroupedPoints[q]].yCoord, (int)(CellSpacing*0.85f));
				LEDIcons[i].display();
				
				GroupedPoints[q] = i;
				
				i = 123456;
				break;
			} //end if()
		} //end for() 

		if(i != 123456) //if its that number, it found an open slot and don't run this
		{
			println("Using: "+IconCounter+"   "+q);
			LEDIcons[IconCounter] = new LEDgraphic(IconCounter, testX, testY, LEDIcons[GroupedPoints[q]].xCoord, LEDIcons[GroupedPoints[q]].yCoord, (int)(CellSpacing*0.85f));
			LEDIcons[IconCounter].display(); 
			
			GroupedPoints[q] = IconCounter; //update GroupSelection array with new value
			IconCounter++; 
			if(IconCounter > SelectSliders[0].Value) { IconCounter--; break;} 
			CountPlaced++;
			//println("Copy Over Maximum, Double check Pixel amount slider");
		}

		println("Ran with "+IconCounter+" at "+testX+" : "+testY);
	} //end for()

	SelectedPointsCounter = CountPlaced; //this makes sure it can't copy greater than user set maximim amount of pixels
	
	if(SelectedPointsCounter > 0) InitMoveGroup();  //now initialize group move function
} //end func()

public void RunGroupSelection()
{
	GroupedPoints = new int[0];   //create empty array of a small size, need to start it at 0 and not 1 because of GroupedPoints.length
	SelectedPointsCounter = 0;

	//  println(SelectBoxXpos+" : "+SelectBoxYpos+" : "+SelectBoxW+" : "+ SelectBoxH);
	for (int q=0; q != IconCounter; q++)
	{
		if (LEDIcons[q].xpos >= SelectBoxXpos && LEDIcons[q].xpos <= SelectBoxXpos+SelectBoxW && LEDIcons[q].ypos >= SelectBoxYpos && LEDIcons[q].ypos <= SelectBoxYpos+SelectBoxH)      
		{
			if(LEDIcons[q].Enabled == true)
			{
				GroupedPoints = expand(GroupedPoints, GroupedPoints.length+1); //expand the array as needed
				println("Icon "+q+" is within bounds");
				GroupedPoints[SelectedPointsCounter] = q;
				SelectedPointsCounter++;
			}
		}
	}
	redraw();
} //end func

//==================================================================================

public void WriteCustomPatchFile()
{
	println("WriteCustomPatchFile()");

	//writes Coordinates in order, Channel Amount tab LED Amount tab Increment/All newline X1 tab Y1 newline X2 tab Y2 etc, no spaces
	strs = new ArrayList(); // reset array list.... 
	int sLines=0;       
	String BuildLine = SelectSliders[0].Value+"\t"+SelectSliders[1].Value+"\t"+SelectSliders[2].Value+"\t"+width+"\t"+height; //create header line

	strs.add(BuildLine);    //write header line to Array list

	for (int i=0;i != IconCounter;i++) 
	//for (int i=0; i != SelectSliders[0].Value; i++)   //only save the amount of pixels that is expected!
	{   
		//again, these are saved as -1, -1 to make up for grid borders
		if (LEDIcons[i].Enabled == true)    BuildLine = (LEDIcons[i].xCoord-1) + "\t"+(LEDIcons[i].yCoord-1);
		else BuildLine = "1\t1"; //if not enabled save point at 1, 1. May overlap but no where else would be easy to store them

		strs.add( BuildLine);
	}   

	String[] lines = new String[strs.size()]; 

	for (int q=0; q != strs.size (); q++)  
	{
		lines[q]=(String) strs.get(q);
	} 
	String[]  ExtensionStripped = split(SelectedFile, '.');
	saveStrings(ExtensionStripped[0]+".txt", lines);  
	println("File Generated: "+ExtensionStripped[0]+".txt");
} //end WriteCustomPatchFile()

//==================================================================================

public void RunRollingAutoSave()
{
	println("Running AutoSave");
	SelectedFile = "autosave/autosavefile-"+rollingSaveNum;
	SaveLayoutFile();
	if (rollingSaveNum == cMaxAutoSaves) rollingSaveNum = 1;
	else rollingSaveNum++;
}

//==================================================================================

public void SaveLayoutFile()
{
	WriteCustomPatchFile();
}
//==================================================================================

public void LoadLayoutFile()
{
	String[] lines = loadStrings(SelectedFile); //divides the lines
	int sLines=0;  
	String[] WorkString = new String[3]; //used to divide the lines into tab

	WorkString = split(lines[0], '\t');

	SelectSliders[0].Value = PApplet.parseInt(WorkString[0]); //channels
	if (SelectSliders[0].Value >  SelectSliders[0].Maximum)  SelectSliders[0].Maximum =  SelectSliders[0].Value;

	SelectSliders[1].Value = PApplet.parseInt(WorkString[1]); //cell size
	if (SelectSliders[1].Value >  SelectSliders[1].Maximum)  SelectSliders[1].Maximum =  SelectSliders[1].Value;  

	SelectSliders[2].Value = PApplet.parseInt(WorkString[2]); //cells
	if (SelectSliders[2].Value >  SelectSliders[2].Maximum)  SelectSliders[2].Maximum =  SelectSliders[2].Value;  

	try 
	{
		int tempW = PApplet.parseInt(WorkString[3]); //grid window width
		int tempH = PApplet.parseInt(WorkString[4]); //grid window height

		//insets get ran in Setup()
		frame.setSize(tempW+(insets.left+insets.right), tempH+(insets.top+insets.bottom));	
		println("Should have resized to "+tempW+" : "+tempH);
	}
	catch(Exception e) { 
		println("Window Size not loaded, may be older file version");
	}

	IconCounter = 0;

	int testX;
	int testY;

	for (int i = 1; i != lines.length; i++)
	{
		WorkString = split(lines[i], '\t');
		testX = (PApplet.parseInt(WorkString[0]) * CellSpacing)+(CellSpacing/2);
		testY = (PApplet.parseInt(WorkString[1]) * CellSpacing)+(CellSpacing/2); 

		//Loads the Coord as +1, +1, to make up for border area with column/row details
		//	saves it as -1, -1, so lowest coord is 1,1 but in software it thinks its 2,2
		LEDIcons[IconCounter] = new LEDgraphic(IconCounter, testX, testY, PApplet.parseInt(WorkString[0])+1, PApplet.parseInt(WorkString[1])+1, (int)(CellSpacing*0.85f));
		IconCounter++;
	}

	UpdateGridCells();
	println("Done Loading");
	redraw();
}//end func


public void FunctionRightClick()
{
    //if over an Icon, open menu, will that work???
    CoordX = round((mouseX / CellSpacing))-OffsetVarX;
    CoordY = round((mouseY / CellSpacing))-OffsetVarY;   

    println("Right Click");

    for (int q=0; q != IconCounter; q++)
    {
      if (LEDIcons[q].over(CoordX, CoordY))
      {
        MenuOpen = true;
        ToolsMenuOpen = false;  

        EnableGroupSelect = false; //make sure group drag or select is not running
        MenuOpenedIcon = q; //used for deleting n such
        MenuXpos = LEDIcons[q].xpos; //set menu position
        MenuYpos = LEDIcons[q].ypos;
        redraw();
        return;
      }
    } //end for()
    //else an icon was not selected, open tool menu
    MenuOpen = false;
    ToolsMenuOpen = true;

    MenuXpos = mouseX; //set menu position
    MenuYpos = mouseY;
    EnableGroupSelect = false;
    redraw();
    return;
}

public void mousePressed()
{
  if (mouseButton == RIGHT)
  {
	FunctionRightClick();
  } 
  else
  {
    //LEFT Mouse Pressed

    //check sliders before rest
    if (GridXScroll.over(mouseX, mouseY))
    {
      println("Selected Scrollbar X");
      SliderSet=0;
      GridXScroll.Dragging=true;
      GlobalDragging=true;
      return;
    }

    if (GridYScroll.over(mouseX, mouseY))
    {
      println("Selected Scrollbar Y");
      SliderSet=1;
      GridXScroll.Dragging=true;
      GlobalDragging=true;
      return;
    }  

    //========= Scrollbar sliders checked, continue ==========

    //factors the translate() values
    int tempMouseX = mouseX - (OffsetVarX*CellSpacing);
    int tempMouseY = mouseY - (OffsetVarY*CellSpacing);

    //check menus first
    if (MenuOpen == true)
    {
      //menu is open 

      //scans menu looking for list number that was pressed
      int i = 0;
      for (i = 0; i != 3; i++)
      {
        if (tempMouseX >= MenuXpos && tempMouseX <= MenuXpos+160 && tempMouseY >= MenuYpos+(i*25) && tempMouseY <= MenuYpos+25+(i*25))    break;
      }

      switch(i) //uses list number to call function
      {
      case 0:
        println("Menu Item 1 - Delete Before - "+MenuOpenedIcon);
        RunRollingAutoSave();
        for (int q = 0; q != MenuOpenedIcon; q++)  LEDIcons[q].Enabled = false;  
        MenuOpen = false;
        break;
      case 1:
        println("Menu Item 2 - Delete After - "+MenuOpenedIcon);
        RunRollingAutoSave();
        for (int q = (MenuOpenedIcon+1); q != IconCounter; q++)  LEDIcons[q].Enabled = false;
        MenuOpen = false;
        break; 
      case 2:
        println("Close Menu");      
        MenuOpen = false;    
        break;   
      case 3:
        println("No item pressed"); 
        MenuOpen = false;          
        break;
      } //end switch
      redraw();
    } //end else
    else if (ToolsMenuOpen == true)
    {    
      //scans menu looking for list number that was presses
      int i = 0;
      for (i = 0; i != 7; i++)
      {
        if (tempMouseX >= MenuXpos && tempMouseX <= MenuXpos+160 && tempMouseY >= MenuYpos+(i*25) && tempMouseY <= MenuYpos+25+(i*25))    break;
      }

      switch(i) //uses list number to call function
      {
      case 0:
        println("Menu Item 1 - Group Select");
        EnableGroupSelect = true;
        SelectPointInit = false;
        MenuOpen = false;
        SelectBoxX = mouseX;
        SelectBoxY = mouseY;
        //      SelectPointInit = true;
        break;
      case 1:
        println("Menu Item 2 - Move Group"); 
        RunRollingAutoSave();
        //updates position for drag handle
        InitMoveGroup();
        break; 
      case 2:
        println("Menu Item 3 - Copy Group");
        RunCopyGroup();
        break;  
      case 3:
        println("Menu Item 4 - Delete Group");
        RunRollingAutoSave();
        for (int q=0; q != SelectedPointsCounter; q++)  LEDIcons[GroupedPoints[q]].Enabled = false;
        break;              
      case 4:
        println("Menu Item 5 - Rotate Group CW");
        RunRollingAutoSave();
        if (DraggingGroup == true) PlacePoint(); //if dragging, place it and pick up again, prevents errors rotating
        RotateGroup(90);
        break;   
      case 5:
        println("Menu Item 6 - Rotate Group CCW");
        RunRollingAutoSave();
        if (DraggingGroup == true) PlacePoint(); //if dragging, place it and pick up again, prevents errors rotating        
        RotateGroup(270);
        break;          
      case 6:
        println("Menu Item 7 - Close Menu");

        break;        
      default:
        println("No item pressed");    
        break;
      } //end switch      
      ToolsMenuOpen = false;   
      return;
    } else if (EnableGroupSelect == true && SelectPointInit == false) 
    {
      SelectBoxX = mouseX;
      SelectBoxY = mouseY;
      SelectPointInit = true;
      println("Selection Box points set");
    } else if (DragToPlaceEnabled == true)
    { 
      println("Starting Drag to Draw");
      RunRollingAutoSave();     
      DragToDraw = true;
      return;
    } else if (EnableGroupSelect == false)  PlacePoint();
  } //end left button else
} //end moused pressed

//==================================================================================

public void mouseDragged()
{
  if (GlobalDragging==true) //only check sliders if one is being dragged
  {
    //println("Dragging "+SliderSet);
    switch(SliderSet)
    {
    case 0:
      GridXScroll.move(mouseX, mouseY); 
      GridXScroll.displayOnGridWindow();

      SelectSliders[3].Value = GridXScroll.Value;

      OffsetVarX = -SelectSliders[3].Value;  
      break;

    case 1:
      GridYScroll.move(mouseX, mouseY); 
      GridYScroll.display();

      SelectSliders[4].Value = GridYScroll.Value;	

      OffsetVarY = -SelectSliders[4].Value;	  
      break;
    }//end switch
  }//end if   
  redraw();
} //end mouseDragging

//==================================================================================

public void mouseReleased()
{
  GlobalDragging = false;

  if (DragToDraw == true)
  {
    DragToDraw = false;	
    println("UnClicked");
    return;
  }

  if (EnableGroupSelect == true && SelectPointInit == true) 
  {
    println("Box Dragging Ended");
    SelectPointInit = false;
    EnableGroupSelect = false;
    RunGroupSelection();
    loop();
  }

  if (DraggingPoint == true)
  {
    CoordX = round((mouseX / CellSpacing)) - OffsetVarX;
    CoordY = round((mouseY / CellSpacing)) - OffsetVarY;   
    DraggingPoint = false; 

    for (int q=0; q != IconCounter; q++)  //LEDIcons.length
    {
      if (LEDIcons[q].over(CoordX, CoordY))
      {
        println("Can Not Drop, Point Already There");
        LEDIcons[SelectedPoint].xpos = (LEDIcons[SelectedPoint].xCoord * CellSpacing)+(CellSpacing/2);
        LEDIcons[SelectedPoint].ypos = (LEDIcons[SelectedPoint].yCoord * CellSpacing)+(CellSpacing/2);
        //  redraw();
        return;
      }
    }

    if (mouseX <= 0 || mouseX >= width  || mouseY <= 0 || mouseY >= height)
    {
      println("Can't Place Off The Window");      
      LEDIcons[SelectedPoint].xpos = (LEDIcons[SelectedPoint].xCoord * CellSpacing)+(CellSpacing/2);
      LEDIcons[SelectedPoint].ypos = (LEDIcons[SelectedPoint].yCoord * CellSpacing)+(CellSpacing/2);   
      //  redraw();    
      return;
    }
    //if dragging and released, refigure point and update object
    int testX = (CoordX * CellSpacing)+(CellSpacing/2);
    int testY = (CoordY * CellSpacing)+(CellSpacing/2); 
    LEDIcons[SelectedPoint] = new LEDgraphic(LEDIcons[SelectedPoint].ChanID, testX, testY, CoordX, CoordY, (int)(CellSpacing*0.85f));
    // redraw(); //was outside this, but only this section reaches the end anyway
  }
  redraw();
} //end mouseReleased()

//==================================================================================

public void keyPressed()
{
  println(PApplet.parseInt(key)); 

  if (key == CODED) 
  {
    if (keyCode == UP)  
    {
      if (OffsetVarY != 0) OffsetVarY++;    

      SelectSliders[3].Value = -OffsetVarY;
      GridYScroll.Value = -OffsetVarY;
    }
    if (keyCode == DOWN) 
    {
      if (OffsetVarY != -(cellWidth-1)) OffsetVarY--;

      SelectSliders[3].Value = -OffsetVarY;
      GridYScroll.Value = -OffsetVarY;
    }
    if (keyCode == RIGHT)  
    {
      if (OffsetVarX != -(cellWidth-1)) OffsetVarX--;
      SelectSliders[4].Value = -OffsetVarX;
      GridXScroll.Value = -OffsetVarX;
    }
    if (keyCode == LEFT) 
    {
      if (OffsetVarX != 0) OffsetVarX++;
      SelectSliders[4].Value = -OffsetVarX;
      GridXScroll.Value = -OffsetVarX;
    }
    if (keyCode == SHIFT) 
    {
      DragToPlaceEnabled =! DragToPlaceEnabled;
      Buttons[6].Selected = DragToPlaceEnabled;
      Buttons[6].display();
    }	
    redraw(); //redraw new Translate
  }//end if coded

  //==================================================================================

  if(key == 32) FunctionRightClick(); //Spacebar simulates right-mouse click
  
  if (key == 26) //ctrl z, UNDO
  {
    println("Running Undo Function - Loading Autosave"); 

    // println("OLD AUTOSAVE ID: "+rollingSaveNum);
    if (rollingSaveNum == 1) rollingSaveNum = cMaxAutoSaves;
    else rollingSaveNum--;
    // println("NEW AUTOSAVE ID: "+rollingSaveNum);

    SelectedFile = "autosave/autosavefile-"+rollingSaveNum+".txt";
    LoadLayoutFile();
  }//end key if
}

class Button
{
  String Label;
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int bWidth;
  int bHeight;
  int ColBG;
  int ColHL;
  int ColOut;  
  boolean Toggle;
  boolean Selected;

  Button(String iLabel, int ixpos, int iypos, int ibWidth, int ibHeight, int iColBG, int iColHL, int iColOut, boolean iToggle, boolean iSelected)
  {
    Label = iLabel; 
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    ColBG = iColBG;
    ColHL = iColHL;
    ColOut = iColOut;
    Toggle = iToggle;
    Selected = iSelected;
  }

  public void display()
  {
    s.strokeWeight(2);
    s.stroke(ColOut);

    if (Selected==true)
    {
      s.fill(ColHL); 
      s.rect(xpos, ypos, bWidth, bHeight, bHeight/4);
    }
    else       
    {
      s.fill(ColBG);
      s.rect(xpos, ypos, bWidth, bHeight, bHeight/4);
    }
    //adjust stroke to on higligh and invert text color

    s.fill(255);
    s.textSize((bHeight/1.75f));
    s.text(Label, xpos+((bWidth - s.textWidth(Label))/2), ypos+((bHeight-(s.textAscent()+s.textDescent()))/2), bWidth, bHeight);

    s.noStroke();
    if (Toggle==false) Selected = false;
  }

  public boolean over(int uX, int uY) 
  {
    if ( uX >= xpos && uX <= xpos+bWidth && uY >= ypos && uY <= ypos+bHeight) 
    {      
      return true;
    } 
    else {
      return false;
    }
  }
} 

//=======================================

class LEDgraphic
{
  int ChanID;
  int IDNumber;
  int xpos; // display rect xposition  
  int ypos ; // displayrect yposition
  int xCoord;
  int yCoord;
  int bSize;
//  color cColor;
  boolean Enabled;

  LEDgraphic(int iChanID, int ixpos, int iypos, int ixCoord, int iyCoord, int ibSize)//, color icColor)
  {
    ChanID=iChanID; 
    xpos = ixpos;
    ypos = iypos;  
    yCoord = iyCoord;
    xCoord = ixCoord;
    bSize = ibSize;
   // cColor = icColor;
   Enabled = true;
  }

  public void display()
  {   
   // float myRed = cColor >> 16 & 0xFF; //convert colors to 8-bit
    //float myGreen = cColor >> 8 & 0xFF;
   // float myBlue = cColor & 0xFF;  

    noStroke();
    fill(LEDIconColor);
    ellipse(xpos, ypos, bSize, bSize);

    //fill((255 - myRed), (255 - myGreen), (255 - myBlue)); //inverts color
    fill(255); //white
    textSize((bSize/1.75f));

    textAlign(CENTER, CENTER); //centers the text on the ellipse
    text((ChanID+1), xpos, ypos-(bSize/10));
    noStroke();
  }

  public void displaySelected()
  {   
    if(Enabled == false) return;
    noStroke();
    fill(255,0,0); //highlight
    ellipse(xpos, ypos, bSize*1.2f, bSize*1.2f);   
  }

  public boolean over(int uX, int uY) 
  { //custom for circles! which are placed on center
    if (uX == xCoord && uY == yCoord) 
    {      
      return true;
    } 
    else {
      return false;
    }
  }

  /*
    if ( uX >= (xpos - (bSize/2)) && uX <= xpos+(bSize/2)&& uY >= (ypos - (bSize/2)) && uY <= ypos+(bSize/2)) 
   {      
   return true;
   } 
   else {
   return false;
   }
   }
   */
}//end class

//=========================================================

class SliderBar
{
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int Slwidth; // single bar Slwidthth
  int Slheight; // rect height
  int Value; 
  int Minimum; 
  int Maximum;
  int BgColor;
  int SlColor;
  int SelColor; 
  int StrokeColor;
  boolean CircleHandle;  
  boolean Dragging;


  //local variables 
  boolean over = false;

  //gotta move the int variables into float variables
  float temp1=Slwidth;
  float temp2=Maximum;
  float temp3=Value;
  float temp4=Minimum;
  float MathVar=0;


  SliderBar(int ixp, int iyp, int iw, int ih, int iValue, int iMin, int iMax, int iBgColor, int iSlColor, int iSelColor, int iStrokeColor, boolean iCircleHandle, boolean iDragging) 
  {   
    xpos = ixp;
    ypos = iyp;  
    Slwidth = iw;
    Slheight = ih;
    Value=iValue;
    Maximum=iMax;
    Minimum=iMin;
    BgColor=iBgColor;
    SlColor=iSlColor;
    SelColor=iSelColor;
    StrokeColor=iStrokeColor;
    CircleHandle=iCircleHandle;     
    Dragging=iDragging;
  }

  public void move(int uX, int uY) 
  {
    //this function handles position of the slider's handle

    temp1 = Slwidth;   
    temp3 = Value;
    temp4 = Minimum;

    MathVar = (((uX-xpos) / temp1)-1) *-1; //creates a scale factor based on mouse position 
    //relative to slider position
    MathVar = Maximum - (MathVar*(Maximum-Minimum));   

    if (MathVar > Maximum) { //check so it doesn't drag off the end
      MathVar=Maximum;
    }

    if (MathVar < Minimum) { //check so it doesn't drag off the end
      MathVar=Minimum;
    }  

    Value=PApplet.parseInt(MathVar); //set the value based on mouse movement
  }


  public void display() 
  {
    s.stroke(StrokeColor);
    s.strokeWeight(1);

    //circle handles are displayed differently so there are if/else statements to set it correctly

    s.fill(BgColor); //covers up the old slider with a blank rectangle, BgColor set in declaration
    s.rect(xpos, ypos, Slwidth, Slheight);   

    if (CircleHandle==true) temp1 = Slwidth -Slheight;
    else temp1=Slwidth-(Slwidth/20);

    temp2=Maximum;  //move variables into float
    temp3=Value;    //or do they don't compute correctly
    temp4 = Minimum;

    float MathVar;

    if (CircleHandle==true)
      MathVar = temp1 / (temp2 - temp4) * (temp3-temp4) + Slheight/2;
    else  
      MathVar = temp1 / (temp2 - temp4) * (temp3-temp4); 
    //MathVar = (SlWidth / (Max / Min) * (Value - Minimum)

    s.fill(color(SelColor)); //draws the Selcolor over the selected area
    s.rect(xpos, ypos, MathVar, Slheight);

    s.fill(SlColor); //places the slider handle
    if (CircleHandle==true)   s.ellipse(xpos+MathVar, ypos+(Slheight/2), Slheight, Slheight); //Circle instead of rectangle
    else   rect((xpos+MathVar), ypos, Slwidth/20, Slheight); //and comment this line for circle
  }

  
   public void displayOnGridWindow() 
  {
    stroke(StrokeColor);
    strokeWeight(1);

    //circle handles are displayed differently so there are if/else statements to set it correctly

    fill(BgColor); //covers up the old slider with a blank rectangle, BgColor set in declaration
    rect(xpos, ypos, Slwidth, Slheight);   

    if (CircleHandle==true) temp1 = Slwidth -Slheight;
    else temp1=Slwidth-(Slwidth/20);

    temp2=Maximum;  //move variables into float
    temp3=Value;    //or do they don't compute correctly
    temp4 = Minimum;

    float MathVar;

    if (CircleHandle==true)
      MathVar = temp1 / (temp2 - temp4) * (temp3-temp4) + Slheight/2;
    else  
      MathVar = temp1 / (temp2 - temp4) * (temp3-temp4); 
    //MathVar = (SlWidth / (Max / Min) * (Value - Minimum)

    fill(color(SelColor)); //draws the Selcolor over the selected area
    rect(xpos, ypos, MathVar, Slheight);

    fill(SlColor); //places the slider handle
    if (CircleHandle==true)   ellipse(xpos+MathVar, ypos+(Slheight/2), Slheight, Slheight); //Circle instead of rectangle
    else   rect((xpos+MathVar), ypos, Slwidth/20, Slheight); //and comment this line for circle
  } 
  
  
  public boolean over(int uX, int uY) 
  {
    if ( uX >= xpos && uX <= xpos+Slwidth && uY >= ypos && uY <= ypos+Slheight) 
    {          
      Dragging=true;
      over = true;
      return true;
    } 
    else {
      over = false;
      return false;
    }
  }
}

//===============================================================================

class VertSliderBar
{
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int Slwidth; // single bar Slwidthth
  int Slheight; // rect height
  int Value; 
  int Minimum; 
  int Maximum;
  int BgColor;
  int SlColor;
  int SelColor; 
  int StrokeColor;
  boolean CircleHandle;  
  boolean Dragging;


  //local variables 
  boolean over = false;

  //gotta move the int variables into float variables
  float temp1=Slwidth;
  float temp2=Maximum;
  float temp3=Value;
  float temp4=Minimum;
  float MathVar=0;


  VertSliderBar(int ixp, int iyp, int iw, int ih, int iValue, int iMin, int iMax, int iBgColor, int iSlColor, int iSelColor, int iStrokeColor, boolean iCircleHandle, boolean iDragging) 
  {   
    xpos = ixp;
    ypos = iyp;  
    Slwidth = iw;
    Slheight = ih;
    Value=iValue;
    Maximum=iMax;
    Minimum=iMin;
    BgColor=iBgColor;
    SlColor=iSlColor;
    SelColor=iSelColor;
    StrokeColor=iStrokeColor;
    CircleHandle=iCircleHandle;     
    Dragging=iDragging;
  }

//  void move () //VERTICLE
    public void move(int uX, int uY) 
  {
    //this function handles position of the slider's handle

    temp1 = Slheight;   
    temp3 = Value;
    temp4 = Minimum;

    MathVar = (((uY-ypos) / temp1)-1) *-1; //creates a scale factor based on mouse position 
    //relative to slider position
    MathVar = Maximum - (MathVar*(Maximum-Minimum));   

    if (MathVar > Maximum) { //check so it doesn't drag off the end
      MathVar=Maximum;
    }

    if (MathVar < Minimum) { //check so it doesn't drag off the end
      MathVar=Minimum;
    }  

    Value=PApplet.parseInt(MathVar); //set the value based on mouse movement
  }


  public void display() //VERTICLE
  {
    stroke(StrokeColor);
    strokeWeight(1);

    //circle handles are displayed differently so there are if/else statements to set it correctly

    fill(BgColor); //covers up the old slider with a blank rectangle, BgColor set in declaration
    rect(xpos, ypos, Slwidth, Slheight);   

    temp1 = Slheight - Slwidth;

    temp2= Maximum;  //move variables into float
    temp3= Value;    //or do they don't compute correctly
    temp4 = Minimum;
    float MathVar;
    MathVar = temp1 / (temp2 - temp4) * (temp3-temp4) + Slwidth/2;
    //MathVar = (SlWidth / (Max / Min) * (Value - Minimum)
    fill(color(SelColor)); //draws the Selcolor over the selected area
    rect(xpos, ypos, Slwidth, MathVar);
    fill(SlColor); //places the slider handle
    ellipse(xpos+(Slwidth/2), ypos+MathVar, Slwidth, Slwidth); //Circle instead of rectangle
  }

  public boolean over(int uX, int uY) 
  {
    if ( uX >= xpos && uX <= xpos+Slwidth && uY >= ypos && uY <= ypos+Slheight) 
    {          
      Dragging=true;
      over = true;
      return true;
    } 
    else {
      over = false;
      return false;
    }
  }
}

//===============================================================================

class TextField
{
  String Label;
  int xpos; // rect xposition  
  int ypos ; // rect yposition
  int bWidth;
  int bHeight;
  int ColBG;
  int ColHL;
  boolean Toggle;
  boolean Selected;

  TextField(String iLabel, int ixpos, int iypos, int ibWidth, int ibHeight, int iColBG, int iColHL, boolean iToggle, boolean iSelected)
  {
    Label = iLabel; 
    xpos = ixpos;
    ypos = iypos;  
    bWidth = ibWidth;
    bHeight = ibHeight;
    ColBG = iColBG;
    ColHL = iColHL;
    Toggle = iToggle;
    Selected = iSelected;
  }

  public void display()
  {
    s.translate(0, 0);
    s.strokeWeight(2);
    s.fill(ColBG);

    if (Selected==true)
    {
      s.stroke(ColHL); 
      s.rect(xpos, ypos, bWidth, bHeight);
    }
    else       
    {
      s.stroke(0, 0, 255);
      s.rect(xpos, ypos, bWidth, bHeight);
    }
    //adjust stroke to on higligh and invert text color
    s.noStroke();
    s.fill(0);
    s.textSize((bHeight/1.75f));
    s.text(Label, xpos+((bWidth - s.textWidth(Label))/2), ypos+((bHeight-(s.textAscent()+s.textDescent()))/2), bWidth, bHeight);

    if (Toggle==false) Selected = false;
  }

  public boolean over(int uX, int uY) 
  {
    if ( uX >= xpos && uX <= xpos+bWidth && uY >= ypos && uY <= ypos+bHeight) 
    {      
      return true;
    } 
    else {
      return false;
    }
  }
} 




  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "NLEDMatrixPatcher" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
